(*
   SNU 4190.310 Programming Languages (Fall 2008)
 
   K-- Interpreter
*)

(* Location : don't mention it *)
signature LOC =
	sig
	type t
	val base : t
	val equal : t -> t -> bool
	val diff : t -> t -> int
	val increase : t -> int -> t
	end
structure Loc : LOC =
	struct
	type t = Location of int
	val base = Location(0)
	fun equal Location(a) Location(b) = a = b
	fun diff Location(a) Location(b) = a - b
	fun increase Location(base) n = Location(base+n)
	end

(* Memory Signature *)
signature MEM = 
	sig
	type 'a t
	exception Not_allocated
	exception Not_initialized
	val empty : 'a t (* get empty memory *)
	val load : 'a t -> Loc.t  -> 'a (* load value : Mem.load mem loc => value *)
	val store : 'a t -> Loc.t -> 'a -> 'a t (* save value : Mem.load mem loc value => mem' *)
	val new : 'a t -> Loc.t * 'a t (* get fresh memory cell : Mem.new mem => (loc, mem') *)
	end

(* Environment Signature *)
signature ENV =
	sig
	type ('a, 'b) t
	exception Not_bound
	val empty : ('a, 'b) t (* get empty environment *)
	val lookup : ('a, 'b) t -> 'a -> 'b (* lookup environment : Env.lookup env key => content *)
	val bind : ('a, 'b) t -> 'a -> 'b -> ('a, 'b) t  (* id binding : Env.bind env key content => env'*)
	end

(* Memory Implementation *)
structure Mem : MEM =
	struct
	exception Not_allocated
	exception Not_initialized
	type 'a content = V of 'a | U
	type 'a t = M of Loc.t * 'a content list
	val empty = M(Loc.base,[])

	val rec replace_nth = 
		fn (h::t) n c => if n = 1 then c::t else h::(replace_nth t (n-1) c)
		 | [] n c => raise Not_allocated

	fun load M(boundary,storage) loc =
		case (List.nth storage ((Loc.diff boundary loc) - 1)) of
		V(v) => v | U => raise Not_initialized

	fun store M(boundary,storage) loc content =
		M(boundary, replace_nth storage (Loc.diff boundary loc) V(content))

	fun new M(boundary,storage) = (boundary,M(Loc.increase boundary 1,U::storage))
	end

(* Environment Implementation *)
structure Env : ENV=
	struct
	exception Not_bound
	type ('a,'b) t = E of 'a -> 'b
	val empty = E(fn x => raise Not_bound)
	fun lookup E(env) id = env id
	fun bind E(env) id loc = E(fn x => if x = id then loc else env x)
	end

(*
 * K-- Interpreter
 *)
signature KMMINUS =
  sig
    exception Error of string
    type id = string
    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | ASSIGN of id * exp          (* assgin to variable *)
      | SEQ of exp * exp            (* sequence *)
      | IF of exp * exp * exp       (* if-then-else *)
      | WHILE of exp * exp          (* while loop *)
      | FOR of id * exp * exp * exp (* for loop *)
      | LETV of id * exp * exp      (* variable binding *)
      | LETF of id * id * exp * exp (* procedure binding *)
      | CALLV of id * exp           (* call by value *)
      | CALLR of id * id            (* call by referenece *)
      | READ of id
      | WRITE of exp
    type program = exp
    type memory
    type env
    type value
    val emptyMemory : memory
    val emptyEnv : env
    val run : memory * env * program -> value
  end

structure K : KMMINUS =
  struct
    exception Error of string

    type id = string
    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | ASSIGN of id * exp
      | SEQ of exp * exp
      | IF of exp * exp * exp
      | WHILE of exp * exp
      | FOR of id * exp * exp * exp
      | LETV of id * exp * exp
      | LETF of id * id * exp * exp
      | CALLV of id * exp
      | CALLR of id * id
      | READ of id
      | WRITE of exp
    type program = exp
    type value =
        Num of int
      | Bool of bool
      | Unit
    type memory = value Mem.t
	type env = (id, env_entry) Env.t
	and  env_entry = Addr of Loc.t | Proc of id * exp * env

    val emptyMemory = Mem.empty
    val emptyEnv = Env.empty

    fun value_int Num n = n
      | value_int Bool _ = raise (Error "Bool type is used as Num type")
      | value_int Unit = raise (Error "Unit type is used as Num type")

    fun value_bool Bool b = b
      | value_bool Num _ = raise (Error "Num type is used as Bool type")
      | value_bool Unit = raise (Error "Unit type is used as Bool type")

    fun value_unit Unit = ()
      | value_unit Num _ = raise (Error "Num type is used as Unit type")
      | value_unit Bool _ = raise (Error "Bool type is used as Unit type")

    fun env_loc e x =
        (case Env.lookup e x of
           Addr l => l
         | Proc _ => raise (Error "not allowed")) handle
          Env.Not_bound => raise (Error "not bound")

    fun env_proc e f = 
        (case Env.lookup e f of 
           Addr _ => raise (Error "not allowed") 
         | Proc p => p) handle 
          Env.Not_bound => raise (Error "not bound")
  
    fun run (mem, env, pgm) = case pgm of _ => raise Error("not implemented") (* implement it! *)
  end
